"""
=============
psimulate CLI
=============

Command line interface for `psimulate`.

.. click:: vivarium_cluster_tools.psimulate.cli:psimulate
   :prog: psimulate
   :show-nested:

"""
from pathlib import Path

import click
from loguru import logger
from vivarium.framework.utilities import handle_exceptions

from vivarium_cluster_tools.psimulate import globals as vct_globals
from vivarium_cluster_tools.psimulate import runner, utilities

shared_options = [
    click.option(
        "--project",
        "-P",
        type=click.Choice(vct_globals.CLUSTER_PROJECTS),
        default=vct_globals.DEFAULT_CLUSTER_PROJECT,
        help="The cluster project under which to run the simulation.",
    ),
    click.option(
        "--queue",
        "-q",
        type=click.Choice(["all.q", "long.q"]),
        default=None,  # dynamically set based on max-runtime
        help="The cluster queue to assign psimulate jobs to. Queue defaults to the "
        "appropriate queue based on max-runtime. long.q allows for much longer "
        "runtimes but there may be reasons to send jobs to that queue even "
        "if they don't have runtime constraints, such as node availability.",
    ),
    click.option(
        "--peak-memory",
        "-m",
        type=int,
        default=3,
        help=(
            "The estimated maximum memory usage in GB of an individual simulate job. "
            "The simulations will be run with this as a limit."
        ),
    ),
    click.option(
        "--max-runtime",
        "-r",
        type=str,
        default="24:00:00",
        help=(
            "The estimated maximum runtime (HH:MM:SS) of the simulation jobs. "
            "By default, the cluster will terminate jobs after 24h regardless of "
            "queue. The maximum supported runtime is 3 days. Keep in mind that the "
            "session you are launching from must be able to live at least as long "
            "as the simulation jobs, and that runtimes by node vary wildly."
        ),
    ),
    click.option(
        "--pdb",
        "with_debugger",
        is_flag=True,
        help="Drop into python debugger if an error occurs.",
    ),
    click.option(
        "--redis",
        type=int,
        default=-1,
        help=(
            f"Number of redis databases to use.  Defaults to a redis instance for every "
            f"{vct_globals.DEFAULT_JOBS_PER_REDIS_INSTANCE} jobs."
        ),
    ),
    click.option("-v", "verbose", count=True, help="Configure logging verbosity."),
    click.option(
        "--no-batch", is_flag=True, help="Don't batch results, write them as they come in."
    ),
    click.option(
        "--no-cleanup",
        is_flag=True,
        hidden=True,
        help="Hidden developer option, if flagged, don't automatically cleanup results directory on failure.",
    ),
]


def pass_shared_options(func):
    # add all the shared options to the command
    for option in shared_options:
        func = option(func)
    return func


@click.group()
def psimulate():
    """A command line utility for running many simulations in parallel.

    You may initiate a new run with the ``run`` sub-command or restart a run
    from where it was stopped by using the ``restart`` sub-command.

    """
    pass


@psimulate.command()
@click.argument("model_specification", type=click.Path(exists=True, dir_okay=False))
@click.argument("branch_configuration", type=click.Path(exists=True, dir_okay=False))
@click.option(
    "--artifact_path",
    "-i",
    type=click.Path(resolve_path=True),
    help="The path to the artifact data file.",
)
@click.option(
    "--result-directory",
    "-o",
    type=click.Path(file_okay=False),
    default=None,
    help="The directory to write results to. A folder will be created in this directory with the same name "
    "as the configuration file.",
)
@pass_shared_options
def run(
    model_specification, branch_configuration, artifact_path, result_directory, **options
):
    """Run a parallel simulation.

    The simulation itself is defined by a MODEL_SPECIFICATION yaml file
    and the parameter changes across runs are defined by a BRANCH_CONFIGURATION
    yaml file.

    The path to the data artifact can be provided as an argument here, in the
    branch configuration, or in the model specification file. Values provided as
    a command line argument or in the branch specification file will override a
    value specified in the model specifications file. If an artifact path is
    provided both as a command line argument and to the branch configuration file
    a ConfigurationError will be thrown.

    If a results directory is provided, a subdirectory will be created with the
    same name as the MODEL_SPECIFICATION if one does not exist. Results will be
    written to a further subdirectory named after the start time of the
    simulation run.

    If a results directory is not provided the base results_directory is taken
    to be /share/costeffectiveness/results.

    """
    utilities.configure_master_process_logging_to_terminal(options["verbose"])
    main = handle_exceptions(runner.main, logger, options["with_debugger"])

    main(
        model_specification,
        branch_configuration,
        artifact_path,
        result_directory,
        {
            "project": options["project"],
            "queue": options["queue"],
            "peak_memory": options["peak_memory"],
            "max_runtime": options["max_runtime"],
        },
        redis_processes=options["redis"],
        no_batch=options["no_batch"],
        no_cleanup=options["no_cleanup"],
    )


@psimulate.command()
@click.argument("results-root", type=click.Path(exists=True, file_okay=False, writable=True))
@pass_shared_options
def restart(results_root, **options):
    """Restart a parallel simulation from a previous run at RESULTS_ROOT.

    Restarting will not erase existing results, but will start workers to
    perform the remaining simulations.  RESULTS_ROOT is expected to be an
    output directory from a previous ``psimulate run`` invocation.

    """
    utilities.configure_master_process_logging_to_terminal(options["verbose"])
    main = handle_exceptions(runner.main, logger, options["with_debugger"])

    main(
        None,
        None,
        None,
        results_root,
        {
            "project": options["project"],
            "queue": options["queue"],
            "peak_memory": options["peak_memory"],
            "max_runtime": options["max_runtime"],
        },
        redis_processes=options["redis"],
        restart=True,
        no_batch=options["no_batch"],
        no_cleanup=options["no_cleanup"],
    )


@psimulate.command()
@click.argument("results-root", type=click.Path(exists=True, file_okay=False, writable=True))
@click.option(
    "--add-draws",
    type=int,
    default=0,
    help="The number of input draws to add to a previous run.",
)
@click.option(
    "--add-seeds",
    type=int,
    default=0,
    help="The number of random seeds to add to a previous run.",
)
@pass_shared_options
def expand(results_root, **options):
    """Expand a previous run at RESULTS_ROOT by adding input draws and/or
    random seeds.

    Expanding will not erase existing results, but will start workers to perform
    the additional simulations determined by the added draws/seeds.
    RESULTS_ROOT is expected to be an output directory from a previous
    ``psimulate run`` invocation.

    """
    utilities.configure_master_process_logging_to_terminal(options["verbose"])
    main = handle_exceptions(runner.main, logger, options["with_debugger"])

    main(
        None,
        None,
        None,
        results_root,
        {
            "project": options["project"],
            "queue": options["queue"],
            "peak_memory": options["peak_memory"],
            "max_runtime": options["max_runtime"],
        },
        redis_processes=options["redis"],
        restart=True,
        expand={"num_draws": options["add_draws"], "num_seeds": options["add_seeds"]},
        no_batch=options["no_batch"],
        no_cleanup=options["no_cleanup"],
    )



"""
==================
Distributed Worker
==================

Custome RQ worker for running vivarium jobs.

"""
import json
import math
import os
import random
from pathlib import Path
from time import sleep, time
from traceback import format_exc
from typing import Mapping

import numpy as np
import pandas as pd
import redis
from loguru import logger
from rq import Queue, get_current_job
from rq.job import JobStatus
from rq.registry import FailedJobRegistry
from rq.worker import Worker

from vivarium_cluster_tools.psimulate import globals as vct_globals
from vivarium_cluster_tools.vipin.perf_counters import CounterSnapshot


def retry_handler(job, *exc_info):
    retries = job.meta.get("remaining_retries", 2)

    if retries > 0:
        retries -= 1
        job.meta["remaining_retries"] = retries
        job.set_status(JobStatus.QUEUED)
        job.exc_info = exc_info
        job.save()
        q = Queue(name=job.origin, connection=job.connection)
        q.enqueue_job(job)
        logger.info(f"Retrying job {job.id}")
    else:
        logger.error(f"Failing job {job.id}")
        # exc_string = Worker._get_safe_exception_string(traceback.format_exception(*exc_info))
        q = Queue(name=job.origin, connection=job.connection)
        failed_queue = FailedJobRegistry(queue=q)
        failed_queue.add(job, exc_string=exc_info)
    return False


class ResilientWorker(Worker):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.acceptable_failure_count = 3

    def work(self, *args, **kwargs):
        worker_ = self.name
        logging_directory = Path(os.environ["VIVARIUM_LOGGING_DIRECTORY"])
        logger.add(logging_directory / (str(worker_) + ".log"), level="DEBUG", serialize=True)
        kwargs["logging_level"] = "DEBUG"

        retries = 0
        while retries < 10:
            try:
                super(ResilientWorker, self).work(*args, **kwargs)
                return
            except redis.exceptions.ConnectionError:
                backoff = random.random() * 60
                logger.error(f"Couldn't connect to redis. Retrying in {backoff}...")
                retries += 1
                sleep(backoff)
        logger.error(f"Ran out of retries. Killing worker")

    def main_work_horse(self, job, queue):
        retries = 0
        while retries < 10:
            try:
                super(ResilientWorker, self).main_work_horse(job, queue)
                return
            except redis.exceptions.ConnectionError:
                backoff = random.random() * 60
                logger.error(f"Couldn't connect to redis. Retrying in {backoff}...")
                retries += 1
                sleep(backoff)
        logger.error(f"Ran out of retries. Killing work horse")

    def fork_work_horse(self, job, queue):
        """Spawns a work horse to perform the actual work and passes it a job."""
        child_pid = os.fork()
        os.environ["RQ_WORKER_ID"] = self.name
        os.environ["RQ_JOB_ID"] = job.id
        if child_pid == 0:
            self.main_work_horse(job, queue)
            os._exit(0)  # just in case
        else:
            self._horse_pid = child_pid
            self.procline("Forked {0} at {1}".format(child_pid, time()))


def worker(parameters: Mapping):
    node = f"{os.environ['SGE_CLUSTER_NAME']}:{os.environ['HOSTNAME']}"
    job = f"{os.environ['JOB_NAME']}: {os.environ['JOB_ID']}:{os.environ['SGE_TASK_ID']}"

    input_draw = parameters["input_draw"]
    random_seed = parameters["random_seed"]
    model_specification_file = parameters["model_specification_file"]
    branch_config = parameters["branch_configuration"]
    logger.info(f"Launching new job {job} on {node}")
    logger.info(
        "Starting job: {}".format(
            (input_draw, random_seed, model_specification_file, branch_config)
        )
    )

    try:
        np.random.seed([input_draw, random_seed])
        worker_ = get_current_job().id

        from vivarium.framework.engine import SimulationContext
        from vivarium.framework.utilities import collapse_nested_dict

        configuration = {}
        run_key = {"input_draw": input_draw, "random_seed": random_seed}

        if branch_config is not None:
            configuration.update(dict(branch_config))
            run_key.update(dict(branch_config))

        input_data_config = {
            "input_draw_number": input_draw,
        }
        if vct_globals.ARTIFACT_PATH_KEY in dict(branch_config).get(
            vct_globals.INPUT_DATA_KEY, {}
        ):
            input_data_config.update(
                {
                    vct_globals.ARTIFACT_PATH_KEY: branch_config[vct_globals.INPUT_DATA_KEY][
                        vct_globals.ARTIFACT_PATH_KEY
                    ]
                }
            )

        configuration.update(
            {
                "run_configuration": {
                    "input_draw_number": input_draw,
                    "run_id": str(worker_) + "_" + str(time()),
                    "results_directory": parameters["results_path"],
                    "run_key": run_key,
                },
                "randomness": {
                    "random_seed": random_seed,
                    "additional_seed": input_draw,
                },
                vct_globals.INPUT_DATA_KEY: input_data_config,
            }
        )

        sim = SimulationContext(model_specification_file, configuration=configuration)
        logger.info("Simulation configuration:")
        logger.info(str(sim.configuration))

        start_time = pd.Timestamp(**sim.configuration.time.start.to_dict())
        end_time = pd.Timestamp(**sim.configuration.time.end.to_dict())
        step_size = pd.Timedelta(days=sim.configuration.time.step_size)
        num_steps = int(math.ceil((end_time - start_time) / step_size))

        start_snapshot = CounterSnapshot()
        event = {"start": time()}  # timestamps of application events
        logger.info("Beginning simulation setup.")
        sim.setup()
        event["simulant_initialization_start"] = time()
        exec_time = {
            "setup_minutes": (event["simulant_initialization_start"] - event["start"]) / 60
        }  # execution event
        logger.info(
            f'Simulation setup completed in {exec_time["setup_minutes"]:.3f} minutes.'
        )

        sim.initialize_simulants()
        event["simulation_start"] = time()
        exec_time["simulant_initialization_minutes"] = (
            event["simulation_start"] - event["simulant_initialization_start"]
        ) / 60
        logger.info(
            f'Simulant initialization completed in {exec_time["simulant_initialization_minutes"]:.3f} minutes.'
        )

        logger.info(f"Starting main simulation loop with {num_steps} time steps")
        sim.run()
        event["results_start"] = time()
        exec_time["main_loop_minutes"] = (
            event["results_start"] - event["simulation_start"]
        ) / 60
        exec_time["step_mean_seconds"] = (
            event["results_start"] - event["simulation_start"]
        ) / num_steps
        logger.info(
            f'Simulation main loop completed in {exec_time["main_loop_minutes"]:.3f} minutes.'
        )
        logger.info(f'Average step length was {exec_time["step_mean_seconds"]:.3f} seconds.')

        sim.finalize()
        metrics = sim.report(print_results=False)
        event["end"] = time()
        end_snapshot = CounterSnapshot()

        do_sim_epilogue(start_snapshot, end_snapshot, event, exec_time, parameters)

        idx = pd.MultiIndex.from_tuples(
            [(input_draw, random_seed)], names=["input_draw_number", "random_seed"]
        )
        output_metrics = pd.DataFrame(metrics, index=idx)
        for k, v in collapse_nested_dict(run_key):
            output_metrics[k] = v
        return output_metrics

    except Exception:
        logger.exception("Unhandled exception in worker")
        job = get_current_job()
        job.meta["root_exception"] = format_exc()
        job.save_meta()
        raise
    finally:
        logger.info(
            "Exiting job: {}".format(
                (input_draw, random_seed, model_specification_file, branch_config)
            )
        )


def do_sim_epilogue(
    start: CounterSnapshot,
    end: CounterSnapshot,
    event: dict,
    exec_time: dict,
    parameters: Mapping,
):
    exec_time["results_minutes"] = (event["end"] - event["results_start"]) / 60
    logger.info(f'Results reporting completed in {exec_time["results_minutes"]:.3f} minutes.')
    exec_time["total_minutes"] = (event["end"] - event["start"]) / 60
    logger.info(f'Total simulation run time {exec_time["total_minutes"]:.3f} minutes.')

    perf_log = logger.add(
        Path(os.environ["VIVARIUM_LOGGING_DIRECTORY"])
        / f'perf.{os.environ["JOB_ID"]}.{os.environ["SGE_TASK_ID"]}.log',
        level="DEBUG",
        serialize=True,
    )
    logger.debug(
        json.dumps(
            {
                "host": os.environ["HOSTNAME"].split(".")[0],
                "job_number": os.environ["JOB_ID"],
                "task_number": os.environ["SGE_TASK_ID"],
                "draw": parameters["input_draw"],
                "seed": parameters["random_seed"],
                "scenario": parameters[
                    "branch_configuration"
                ],  # assumes leaves of branch config tree are scenarios
                "event": event,
                "exec_time": exec_time,
                "counters": (end - start).to_dict(),
            }
        )
    )
    logger.remove(perf_log)



"""
===================
psimulate Utilities
===================

Utilities for psimulate runs.

"""
import os
import sys
from datetime import datetime
from pathlib import Path
from shutil import rmtree
from typing import Dict, List, Sequence, Tuple

from loguru import logger

from vivarium_cluster_tools import utilities as vct_utils
from vivarium_cluster_tools.psimulate import globals as vct_globals


def get_drmaa() -> object:
    try:
        import drmaa
    except (RuntimeError, OSError):
        if "SGE_CLUSTER_NAME" in os.environ:
            os.environ["DRMAA_LIBRARY_PATH"] = "/opt/sge/lib/lx-amd64/libdrmaa.so"
            import drmaa
        else:
            drmaa = object()
    return drmaa


def add_logging_sink(sink, verbose: int, colorize=False, serialize=False):
    message_format = (
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> "
        "- <level>{message}</level>"
    )
    if verbose == 0:

        def quiet_filter(record):
            return record.get("extra", {}).get("queue", None) == "all"

        logger.add(
            sink,
            colorize=colorize,
            level="INFO",
            format=message_format,
            filter=quiet_filter,
            serialize=serialize,
        )
    elif verbose == 1:
        logger.add(
            sink, colorize=colorize, level="INFO", format=message_format, serialize=serialize
        )
    elif verbose >= 2:
        logger.add(
            sink, colorize=colorize, level="DEBUG", format=message_format, serialize=serialize
        )


def configure_master_process_logging_to_terminal(verbose: int):
    logger.remove(0)  # Clear default configuration
    add_logging_sink(sys.stdout, verbose, colorize=True)


def configure_master_process_logging_to_file(output_directory: Path):
    master_log = output_directory / "master.log"
    serial_log = output_directory / "master.log.json"
    add_logging_sink(master_log, verbose=2)
    add_logging_sink(serial_log, verbose=2, serialize=True)


def get_output_directory(
    model_specification_file: str = None, output_directory: str = None, restart: bool = False
) -> Path:
    if restart:
        output_directory = Path(output_directory)
    else:
        root = (
            Path(output_directory)
            if output_directory
            else Path(vct_globals.DEFAULT_OUTPUT_DIRECTORY)
        )
        launch_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        model_specification_name = Path(model_specification_file).stem
        output_directory = root / model_specification_name / launch_time
    return output_directory


def set_permissions(output_dir: Path):
    """Call to achieve side effect of relaxing permissions to 775
    on output dir and the parent"""
    permissions = 0o775
    output_dir.parent.chmod(permissions)
    output_dir.chmod(permissions)


def setup_directories(
    model_specification_file: str, result_directory: str, restart: bool, expand: bool
) -> Tuple[Path, Dict[str, Path]]:
    output_directory = get_output_directory(
        model_specification_file, result_directory, restart
    )

    if restart and not expand:
        command = "restart"
        launch_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    elif restart and expand:
        command = "expand"
        launch_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    else:
        command = "run"
        launch_time = output_directory.stem

    logging_directory = output_directory / "logs" / f"{launch_time}_{command}"

    logging_dirs = {
        "main": logging_directory,
        "sge": logging_directory / "sge_logs",
        "worker": logging_directory / "worker_logs",
    }

    vct_utils.mkdir(output_directory, exists_ok=True, parents=True)
    for d in logging_dirs.values():
        vct_utils.mkdir(d, parents=True)

    return output_directory, logging_dirs


def get_hostname() -> str:
    return os.environ.get(vct_globals.CLUSTER_ENV_HOSTNAME)


def exit_if_on_submit_host(name: str):
    if vct_globals.SUBMIT_HOST_MARKER in name:
        raise RuntimeError("This tool must not be run from a submit host.")


def chunks(l: Sequence, n: int):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i : i + n]


def parse_package_version(s: str) -> Tuple[str, str]:
    if (
        "no version control" in s
    ):  # installed from non-git controlled source code in editable mode
        s = s.split("(")[1].split(")")[0]  # yields <package>==<version>

    packaged = s.split("==")
    if len(packaged) == 2:  # not installed from source code, e.g., <package>==<version>
        package = packaged[0]
        version = packaged[1]
    else:  # installed from source code in editable mode so more parsing required
        sourced = s.split("=")
        package = sourced[-1]
        version = sourced[0].split("@")[-1].replace("#egg", "")  # version here is git hash
    return package, version


def convert_pip_list_to_dict(pf_list: List[str]) -> Dict:
    return {p: v for p, v in [parse_package_version(s) for s in pf_list]}


def compare_environments(current: Dict, original: Dict):
    differences = []

    current_packages = set(current.keys())
    original_packages = set(original.keys())

    new_packages = current_packages.difference(original_packages)
    if new_packages:
        differences.append(
            f"The current environment contains the following packages not present in the original "
            f"environment: {new_packages}."
        )

    missing_packages = original_packages.difference(current_packages)
    if missing_packages:
        differences.append(
            f"The current environment is missing the following packages present in the original "
            f"environment: {missing_packages}."
        )

    differing_versions = []
    for p in current_packages.intersection(original_packages):
        if current[p] != original[p]:
            differing_versions.append(f"{p}: {original[p]} -> {current[p]}")

    if differing_versions:
        differences.append(
            f"Different versions found in current and original environment for the following "
            f"packages: {differing_versions}."
        )

    if differences:
        differences = "\n".join(differences)
        raise ValueError(
            f"Differences found between environment used for original run and current environment. "
            f"In order to successfully run, you should make a new environment using the requirements.txt "
            f"file found in the output directory. Differences found as follows: {differences}."
        )


def validate_environment(output_dir: Path):
    original_environment_file = output_dir / "requirements.txt"

    import subprocess

    pip_list_proc = subprocess.Popen(
        ["pip", "list", "--format=freeze"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out, err = pip_list_proc.communicate()
    current_environment_list = out.decode().strip().split("\n")

    if not original_environment_file.exists():  # original run
        with original_environment_file.open("w") as f:
            f.write("\n".join(current_environment_list))
    else:  # compare with original
        with original_environment_file.open("r") as f:
            original_environment_list = [line.replace("\n", "") for line in f]
        current_environment = convert_pip_list_to_dict(current_environment_list)
        original_environment = convert_pip_list_to_dict(original_environment_list)
        compare_environments(current_environment, original_environment)
        logger.info(
            "Validation of environment successful. All pip installed packages match "
            "original versions. Run can proceed."
        )


def check_for_empty_results_dir(output_dir: Path):
    """Remove the results directory including runner and worker logs if the simulation produced no results (i.e.,
    it failed)."""
    if not (output_dir / "output.hdf").exists():
        rmtree(output_dir)



"""
===================
psimulate Constants
===================

Shared constants for psimulate runs.
"""

# Cluster projects under the purview of Simulation Science
CLUSTER_PROJECTS = [
    "proj_cost_effect",
    "proj_cost_effect_diarrhea",
    "proj_cost_effect_dcpn",
    "proj_cost_effect_conic",
    "proj_csu",
]
DEFAULT_CLUSTER_PROJECT = "proj_cost_effect"

DEFAULT_OUTPUT_DIRECTORY = "/share/costeffectiveness/results"

DEFAULT_JOBS_PER_REDIS_INSTANCE = 1000
DEFAULT_THREADS_PER_JOB = 1

# Cluster specific parameters
ALL_Q_MAX_RUNTIME_HOURS = 3 * 24
LONG_Q_MAX_RUNTIME_HOURS = 16 * 24

CLUSTER_ENV_HOSTNAME = "HOSTNAME"
SUBMIT_HOST_MARKER = "-submit-"


# Configuration key constants
INPUT_DATA_KEY = "input_data"
ARTIFACT_PATH_KEY = "artifact_path"



"""
===============
vipin Utilities
===============

Mainly logging.

"""
import sys

from loguru import logger


def add_logging_sink(sink, verbose, colorize=False, serialize=False):
    message_format = (
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> "
        "- <level>{message}</level>"
    )
    if verbose == 0:
        logger.add(
            sink, colorize=colorize, level="INFO", format=message_format, serialize=serialize
        )
    elif verbose >= 1:
        logger.add(
            sink, colorize=colorize, level="DEBUG", format=message_format, serialize=serialize
        )


def configure_master_process_logging_to_terminal(verbose):
    logger.remove()  # Clear default configuration
    add_logging_sink(sys.stdout, verbose, colorize=True)
